%%%%%%%%%%%%%%%%%%%%%%%%%%%%% -*- Mode: Prolog -*- %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% lite-refine96-problem.pl --- Illustrate problem with lite in refine96 file
%% Author          : Ben Moszkowski
%% Created On      : Wed Mar 10 17:53:41 1999
%% Last Modified By: Ben Moszkowski
%% Last Modified On: Wed Mar 10 18:56:23 1999
%% Update Count    : 26
%% Status          : Unknown, Use with caution!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-op(65,xfy,['<~']).

:- verbose(off).

% Satisfy
sat(X) :- verbose,ex(not(X)).
sat(X) :- ex(not(X)),nl,diag.

% Prove (verify)
pr(X) :- verbose,ex(X).
pr(X) :- ex(X),nl,diag.


/*
 I use padded temporal assignment P<~Q which specifies that
 P remains unchanged until the last state where it equals Q's
 initial value.
*/

define('[m]'(Q), #((more->Q))).
define(padded(Q), ('[m]'(Q);'[m]'(not(Q)))).
define(P<~Q, ((fin(P)=Q),padded(P))).
/*
define(P<~Q, ((fin(P)=Q),padded(Q))).
  is wrong. and fixed in these examples.
   Thu Mar 18 14:48:12 GMT+0900 1999
 */


define(problem,
    ((x1,(s1 <~ true),(stable(c1))) &
     (x3,(c1 <~ true),(stable(s1))) &
     (x4,(c1 <~ false),(stable(s1))) &
     (x5,(s1 <~ false),(stable(c1))) &
     (x6,(s1 <~ true),(stable(c1))))
).

define(ok,
    ((x1,s1_up) &
     (x3,s1_down) &
     (x4,c1_up) &
     (x5,c1_down) &
     (x6,s1_up))
).

define(exclude,[]((
    x1,~x3,~x4,~x5,~x6;
    ~x1,x3,~x4,~x5,~x6;
    ~x1,~x3,x4,~x5,~x6;
    ~x1,~x3,~x4,x5,~x6;
    ~x1,~x3,~x4,~x5,x6))
).

define(not_ok,
    (((x1),s1_up) &
     ((x3),c1_up) &
     ((x4),c1_down) &
     ((x5),s1_down) &
     ((x6),s1_up))).

define(not_ok1,(exclude,
    ((keep(x1),s1_up) &
     (keep(x3),c1_up) &
     (keep(x4),c1_down) &
     (keep(x5),s1_down) &
     (keep(x6),s1_up)))).

define(spec,(~s1,~c1, 
    (true & (empty,s1,c1) & skip & (empty,s1,~c1)
          & skip & (empty,s1,c1) & true))).


define(s1_up,
   (s1<~true,stable(c1))).

define(s1_down,
   (s1<~false,stable(c1))).

define(c1_up,
   (c1<~true,stable(s1))).

define(c1_down,
   (c1<~false,stable(s1))).

% This should be valid but Lite says it is not.
test1:- pr(((~s1,~c1,not_ok)  ->  '[a]'(((c1, <> ~c1,fin(c1))  ->  <> ~s1)))).

% Both examples work correct when "ok" is used instead of "problem":
test2:- sat((~s1,~c1, not_ok,
    (true & (empty,s1,c1) & skip & (empty,s1,~c1)
          & skip & (empty,s1,c1) & true))).
test3:- pr(((~s1,~c1,not_ok)  ->  '[a]'(((c1, <> ~c1,fin(c1))  ->  <> ~s1)))).

test4 :- sat((~s1,~c1, not_ok,
    (true & (empty,s1,c1) & skip & (empty,s1,~c1)
          & skip & (empty,s1,c1) & true))).

% This should be valid but Lite says it is not.
ok1:- pr(((~s1,~c1,problem)  ->  '[a]'(((c1, <> ~c1,fin(c1))  ->  <> ~s1)))).

% Both examples work correct when "ok" is used instead of "problem":
ok2:- sat((~s1,~c1, problem,
    (true & (empty,s1,c1) & skip & (empty,s1,~c1)
          & skip & (empty,s1,c1) & true))).
ok3:- pr(((~s1,~c1,problem)  ->  '[a]'(((c1, <> ~c1,fin(c1))  ->  <> ~s1)))).

ok4 :- sat((~s1,~c1, problem,exlude,
    (true & (empty,s1,c1) & skip & (empty,s1,~c1)
          & skip & (empty,s1,c1) & true))).

